require 'oj'
require 'rest-client'


Oj.default_options = { mode: :compat }

module Esteem

  # Cribbed from geronime's "es-reindex" script: https://github.com/geronime/es-reindex
  class Reindex
    attr_reader :success


    # Perform a new reindex job. Success or failure will be reported by the
    # read-only "success" attribute.
    #
    # @option [String] :source URL for source index and type
    # @option [String] :destination URL for destination index
    # @option [Integer] :frame Scroll window size
    # @option [Boolean] :update Set to update existing documents
    def initialize source:, destination:, frame:, update:
      src, dst = source, destination
      @success = false

      type, src = src.reverse.split('/', 2).map(&:reverse)

      surl, durl, sidx, didx = '', '', '', ''
      [[src, surl, sidx], [dst, durl, didx]].each{|param, url, idx|
        if param =~ %r{^(.*)/(.*?)$}
          url.replace $1
          idx.replace $2
        else
          url.replace 'http://127.0.0.1:9200'
          idx.replace param
        end
      }

      puts "Copying '%s/%s/%s' to '%s/%s'" % [ surl, sidx, type, durl, didx ]

      scount = retried_request :get, "#{surl}/#{sidx}/#{type}/_count?q=*"
      scount = Oj.load(scount)['count'].to_i

      puts "%d documents" % scount

      if scount == 0
        puts "Hey, this index is empty! Skipping..."
        @success = true
        return
      end

      t, done = Time.now, 0
      shards = retried_request :get, "#{surl}/#{sidx}/_count?q=*"
      shards = Oj.load(shards)['_shards']['total'].to_i
      scan = retried_request(:get, "#{surl}/#{sidx}/#{type}/_search?search_type=scan&scroll=10m&size=#{frame / shards}")
      scan = Oj.load scan
      scroll_id = scan['_scroll_id']
      total = scan['hits']['total']
      printf "    %u/%u (%.1f%%) done.\r", done, total, 0

      bulk_op = update ? "index" : "create"

      loop do
        data = retried_request(:get, "#{surl}/_search/scroll?scroll=10m&scroll_id=#{scroll_id}")
        data = Oj.load data
        break if data['hits']['hits'].empty?

        scroll_id = data['_scroll_id']
        bulk = ''

        data['hits']['hits'].each do |doc|
          bulk << %Q({"#{bulk_op}": {"_index" : "#{didx}", "_id" : "#{
              doc['_id']}", "_type" : "#{doc['_type']}"}}\n)
          bulk << Oj.dump(doc['_source']) + "\n"
          done += 1
        end

        unless bulk.empty?
          bulk << "\n" # empty line in the end required
          retried_request :post, "#{durl}/_bulk", bulk
        end

        eta = total * (Time.now - t) / done
        printf "    %u/%u (%.1f%%) done in %s, E.T.A.: %s.\r",
          done, total, 100.0 * done / total, tm_len(Time.now - t), t + eta
      end

      printf "#{' ' * 80}\r    %u/%u done in %s.\n",
        done, total, tm_len(Time.now - t)



      tries = 0
      dcount = -1

      puts "Waiting for documents to appear in index..."
      until total <= dcount
        dcount = retried_request :get, "#{durl}/#{didx}/#{type}/_count"
        dcount = Oj.load(dcount)['count'].to_i
        tries += 1

        if tries > 30
          puts "Too many tries!"
          @success = false
          return
        end

        sleep 15
      end

      if done != dcount
        puts "Incomplete done! %d != %d" % [ done, dcount ]
        @success = false
        return
      end

      @success = true
    end



  private

    def tm_len l
      t = []
      t.push l/86400; l %= 86400
      t.push l/3600;  l %= 3600
      t.push l/60;    l %= 60
      t.push l
      out = sprintf '%u', t.shift
      out = out == '0' ? '' : out + ' days, '
      out += sprintf('%u:%02u:%02u', *t)
      out
    end


    def retried_request method, url, data=nil
      while true
        begin
          return data ?
            RestClient.send(method, url, data) :
            RestClient.send(method, url)
        rescue RestClient::ResourceNotFound # no point to retry
          return nil
        rescue => e
          warn "\nRetrying #{method.to_s.upcase} ERROR: #{e.class} - #{e.message}"
        end
      end
    end

  end
end